from supadef.platform import run_server_task
from supadef.util import run_step
from .version import VERSION
from .push import push_project, set_project_env
import subprocess
import json
import requests
from typer import Typer, echo
import typer
from .config import TIMEOUT_SECONDS, SERVICE_ENDPOINT, LOCAL_CREDS_PATH
from tabulate import tabulate
from yaspin import yaspin
from .network import GET, POST, get_auth_headers, GET_TEXT

app = Typer()


def execute_bash_command(cmd):
    print(cmd)
    process = subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    if process.returncode != 0:
        raise subprocess.CalledProcessError(
            process.returncode, cmd, output=stdout, stderr=stderr)
    return stdout.decode().strip()


@app.command()
def connect():
    """check that you can securely connect to the supadef platform"""
    with yaspin(text="Connecting to supadef platform", color="yellow") as sp:
        json = GET('/cli/connect')
        sp.text = f'Connected [{json}]'
        sp.ok("✅ ")


@app.command()
def create(project_name: str):
    """create a new project"""

    with yaspin(text="Creating project", color="yellow") as sp:
        json = POST('/projects/create', {
            'name': project_name
        })
        sp.text = f"Project created: '{project_name}'"
        sp.ok("✅ ")


@app.command()
def projects():
    """list your projects"""
    with yaspin(text="Getting projects", color="yellow") as sp:
        json = GET('/projects/json')

        sp.text = f'Done'
        sp.ok("✅ ")

        headers = ['name', 'created_at']
        table = [[p[x] for x in headers] for p in json]
        echo(tabulate(table, headers=headers))


@app.command(name='sd_ignore')
def sd_ignore():
    """View the default .sd_ignore file (.gitignore style - exclude files from supadef build)"""
    with yaspin(text="Getting default .sd_ignore file", color="yellow") as sp:
        t = GET_TEXT('/defaults/.sd_ignore')

        sp.text = f'Done'
        sp.ok("✅ ")
        echo(t)


@app.command()
def push(project_name: str, path_to_code: str, debug: bool = typer.Option(False, "--debug", help="Enable debug mode")):
    """push your code to a project"""
    push_project(project_name, path_to_code, debug)


@app.command(name='set_env')
def set_env(project_name: str, path_to_env_file: str, debug: bool = typer.Option(False, "--debug", help="Enable debug mode")):
    """Securely upload an environment file (.env) to your project"""
    set_project_env(project_name, path_to_env_file, debug)


@app.command(name='run_server_task')
def _run_server_task(project: str,
                     function: str,
                     args: str):
    """run your function in the cloud"""

    args_dict = json.loads(args)

    def run_it(log):
        run_server_task(project, function, args_dict)

    result = run_step(f'[@server_task] run {function}', run_it)
    return result


@app.command()
def logs(run_id: str):
    """
    get a function's run logs, for a particular run
    """
    with yaspin(text="Getting logs...", color="yellow") as sp:
        run_url = link('supadef logs', run_id=run_id)
        response = requests.get(
            run_url, headers=get_auth_headers(), timeout=TIMEOUT_SECONDS)

        if response.status_code == 200:
            sp.text = f'Got logs'
            sp.ok("✅ ")
            pretty_json = json.dumps(response.json(), indent=4)
            print(pretty_json)
        else:
            sp.text = 'Something went wrong'
            sp.fail()
            print(response.status_code)
            print(response.json())


@app.command()
def destroy(project_name: str):
    """destroy a project"""
    body = {
        'name': project_name
    }
    response = requests.delete(link('supadef destroy'), headers=get_auth_headers(
    ), json=body, timeout=TIMEOUT_SECONDS)
    print(response.status_code)
    print(response.json())


@app.command()
def about():
    """basic info about the running version of the CLI"""
    print(f'v{VERSION}')
    print(f'Service endpoint: {SERVICE_ENDPOINT}')
    print(f'Looking for credentials at: {LOCAL_CREDS_PATH}')


# @app.command()
# def spinner():
#     with yaspin(text="Connecting to platform", color="yellow") as spinner:
#         time.sleep(2)  # time consuming code
#         spinner.ok("✅ ")
#
#     with yaspin(text="Creating project record", color="yellow") as spinner:
#         time.sleep(2)  # time consuming code
#         spinner.ok("✅ ")
#
#     with yaspin(text="Provisioning project resources", color="yellow") as spinner:
#         time.sleep(2)  # time consuming code
#         spinner.ok("✅ ")
#
#     with yaspin(text="Deploying latest", color="yellow") as spinner:
#         time.sleep(2)  # time consuming code
#         spinner.ok("✅ ")
#
#     # with yaspin(text="Creating project record", color="yellow") as spinner:
#     #     time.sleep(2)  # time consuming code
#     #
#     #     success = randint(0, 1)
#     #     if success:
#     #         spinner.ok("✅ ")
#     #     else:
#     #         spinner.fail("💥 ")


if __name__ == "__main__":
    app()



from enum import Enum
import traceback
from yaspin import yaspin


class FailMode(Enum):
    EXIT = 1
    THROW_ERROR = 2


def serialize_exception(exception):
    return {
        'type': type(exception).__name__,
        'message': str(exception),
        'traceback': ''.join(traceback.format_exception(None, exception, exception.__traceback__))
    }


def run_step(step_name: str, f: callable, fail_mode: FailMode = FailMode.EXIT):
    """
    Run a given step, defined in the function f.

    To make it possible for the function f to 'print' outputs, it must be of the form:

    f(log: callable)

    then, you can log outputs with

    log('your message')
    """

    with yaspin(text=f"Running: [{step_name}]", color="yellow") as sp:
        def log(msg):
            sp.write(msg)

        try:
            out = f(log)
            sp.text = ''
            sp.ok(f"Done ✅: [{step_name}]")
            return out
        except Exception as e:
            sp.text = str(e)
            sp.fail(f"Error ❌: [{step_name}]")
            if fail_mode == FailMode.EXIT:
                exit()
            if fail_mode == FailMode.THROW_ERROR:
                raise e



from typing import Callable, Optional, Union

from supadef.network import POST


def run_server_task(project_name: str, func: Union[str, Callable], args: dict, module: Optional[str] = None):
    func_name = func
    if callable(func):
        func_name = func.__name__

    json = POST(f'/run/server_task', {
        'project_name': project_name,
        'func_name': func_name,
        'args': args,
        'module': module
    })
    return json



from .config import SERVICE_ENDPOINT

SLUG_TO_PATTERN = {
    'user': '/{username}',
    'user > billing': '/{username}/billing',
    'user > account': '/{username}/account',
    'user > project': '/{username}/{project}',
    'user > project > functions > run': '/{username}/{project}/{function}/{run_id}',
    'user > project > danger zone': '/{username}/{project}/danger_zone',
    'user > project > env': '/{username}/{project}/env',
    'docs > cli > supadef connect': '',
    'site > waitlist': '/get_api_key',
    'supadef connect': '/email',
    'supadef projects': '/projects',
    'supadef create': '/project',
    'supadef logs': '/fn/logs/{run_id}',
    'supadef push': '/project/{project}/upload_package',
    'supadef run': '/run',
    'supadef destroy': '/project',
    'supadef set_env': '/project/{project}/set_env'
}


# this is silly? yagni? or nah?
# perhaps if it's constrained to routes for user-specific data
# strictly GET requests (or ws 'GET's). no other operations encoded here
def route(slug: str, loader=False) -> str:
    if slug not in SLUG_TO_PATTERN.keys():
        raise ValueError(f'no such slug: {slug}')

    pattern = SLUG_TO_PATTERN[slug]
    if loader:
        pattern = pattern + '/content_html'

    return pattern


def link(slug: str,
         loader=False,
         username: str = None,
         project: str = None,
         function: str = None,
         run_id: str = None,
         service_endpoint: str = SERVICE_ENDPOINT) -> str:
    subs = {
        'username': username,
        'project': project,
        'function': function,
        'run_id': run_id
    }
    pattern = route(slug, loader).format(**subs)

    return service_endpoint + pattern



import os

TIMEOUT_SECONDS = 3 * 60
SERVICE_ENDPOINT = 'https://supadef.com'
if 'SUPADEF_SERVICE_ENDPOINT' in os.environ:
    SERVICE_ENDPOINT = os.environ['SUPADEF_SERVICE_ENDPOINT']

LOCAL_CREDS_PATH = '~/.supadef/credentials.yml'



from supadef.credentials import parse_credentials


def test_parse_credentials():
    creds = parse_credentials('tests/test_credentials.yml')
    assert creds
    assert 'default' in creds
    assert ':' in creds['default']

    key_id, key = creds['default'].split(':')

    assert key_id == 'my_id'
    assert key == 'asdfasdfasdf'



import orjson
import requests

from supadef.credentials import get_api_key
from .config import SERVICE_ENDPOINT, TIMEOUT_SECONDS


def get_auth_headers():
    api_key = get_api_key(profile=None)

    headers = {
        "Authorization": f"{api_key}",
        "Content-Type": "application/json"
    }
    return headers


def GET(route: str, params: dict = None, debug=False) -> dict:
    response = requests.get(f'{SERVICE_ENDPOINT}{route}',
                            headers=get_auth_headers(),
                            timeout=TIMEOUT_SECONDS,
                            params=params)
    json = response.json()
    if not response.status_code == 200:
        error_msg = json['detail']
        raise ValueError(error_msg)
    return json


def POST(route: str, body: dict, handle_error=None) -> dict:
    response = requests.post(f'{SERVICE_ENDPOINT}{route}',
                             headers=get_auth_headers(),
                             timeout=TIMEOUT_SECONDS,
                             data=orjson.dumps(body))
    json = response.json()
    if not response.status_code == 200:
        error_msg = json['detail']
        raise ValueError(error_msg)
    return json


def GET_TEXT(route: str) -> dict:
    response = requests.get(f'{SERVICE_ENDPOINT}{route}')
    t = response.text
    if not response.status_code == 200:
        raise ValueError(t)
    return t



import os
import shutil
import tempfile
import zipfile
from typing import List
from pathspec import PathSpec
from pathspec.patterns import GitWildMatchPattern


def create_tempdir(key, nuke_existing=True):
    """
    Create a clean temporary directory.
    """
    # get the path that we should do our work in
    cwd = os.path.join(tempfile.gettempdir(), 'com.supadef.tempdir', key)
    # nuke the working directory, if it exists
    if nuke_existing and os.path.exists(cwd):
        # https://stackoverflow.com/questions/6996603/how-can-i-delete-a-file-or-folder-in-python
        shutil.rmtree(cwd)
    # create it
    os.makedirs(cwd)
    return cwd


def zip_directory(directory_path, zip_filename):
    # Create a ZIP file
    with zipfile.ZipFile(zip_filename, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, _, files in os.walk(directory_path):
            for file in files:
                file_path = os.path.join(root, file)
                arcname = os.path.relpath(file_path, directory_path)
                zipf.write(file_path, arcname=arcname)


def get_non_ignored_files(directory: str, gitignore_content: str) -> List[str]:
    # Create a PathSpec object from the gitignore content
    spec = PathSpec.from_lines(
        GitWildMatchPattern, gitignore_content.splitlines())

    non_ignored_files = []
    for root, _, files in os.walk(directory):
        for file in files:
            file_path = os.path.relpath(
                os.path.join(root, file), directory)
            if not spec.match_file(file_path):
                non_ignored_files.append(file_path)

    return non_ignored_files


def copy_dir(source_dir: str, dest_dir: str, gitignore_content: str):
    # Get the list of files to copy
    files_to_copy = get_non_ignored_files(source_dir, gitignore_content)

    # Ensure the destination directory exists
    os.makedirs(dest_dir, exist_ok=True)

    # Copy each non-ignored file
    for file_path in files_to_copy:
        source_file = os.path.join(source_dir, file_path)
        dest_file = os.path.join(dest_dir, file_path)

        # Create the destination directory if it doesn't exist
        os.makedirs(os.path.dirname(dest_file), exist_ok=True)

        # Copy the file
        shutil.copy2(source_file, dest_file)
        print(f"Copied: {file_path}")

    print(f"Copied {len(files_to_copy)} files from {source_dir} to {dest_dir}")



from . import platform


def server_task():
    """
    Make a function runnable as a task on your web servers.
    :return:
    """
    def g(f):
        # intentionally do nothing :)
        return f
    return g


def endpoint(method='GET'):
    """
    Make a function runnable as a scalable, cloud-hosted API Endpoint.
    :return:
    """
    def g(f):
        # intentionally do nothing :)
        return f
    return g



import os
import requests
from .network import GET, GET_TEXT, POST
from .util import run_step, serialize_exception, FailMode
from .filesystem import create_tempdir, zip_directory, get_non_ignored_files, copy_dir


def zip_directory_in_isolated_tempdir_v2(path_to_code: str) -> str:
    # create an isolated directory to build the package
    work_dir = create_tempdir('supadef_packages')

    def read_gitignore():
        gitignore_path = os.path.join(path_to_code, '.gitignore')

        # Check if .gitignore file exists
        if os.path.exists(gitignore_path):
            with open(gitignore_path, 'r') as file:
                content = file.read()
            return content
        return ''
    # pull down the default .sd_ignore
    default_sd_ignore = GET_TEXT('/defaults/.sd_ignore')

    # read in the user's .gitignore
    user_git_ignore = read_gitignore()

    copy_dir(path_to_code, work_dir, default_sd_ignore + '\n' + user_git_ignore)

    # wire up the full path to the package.zip file
    zip_filename = "package.zip"
    path_to_package_zip = os.path.join(work_dir, zip_filename)
    # zip up the code in the working dir, which has the client code. place output in that dir
    zip_directory(work_dir, path_to_package_zip)
    return path_to_package_zip


def __upload_file_via_presigned_url(file_path, platform_url, project_name, debug):
    def step1_get_upload_url(log):
        r = GET(platform_url,
                params={'project_name': project_name})
        if debug:
            log(f"GET('{platform_url}') -> {r}")
        if not 'url' in r:
            raise ValueError('Could not get upload URL - missing url')
        if not 'fields' in r:
            raise ValueError('Could not get upload URL - missing fields')
        return r

    presigned_post_data = run_step('Get upload URL', step1_get_upload_url)

    def step2_upload_file(log):
        def upload_file_with_redirect_handling(url, fields, file_path):
            max_retries = 3  # Number of retries
            for attempt in range(max_retries):
                try:
                    # Open the file to upload
                    with open(file_path, 'rb') as file:
                        if debug:
                            log(f'[start] POST: {url}')
                        response = requests.post(
                            url,
                            data=fields,
                            files={'file': (fields['key'], file)},
                            allow_redirects=False,  # Disable automatic redirect handling
                        )
                        if debug:
                            log(f'[response] POST: {url} | {response}')

                    # Check if a redirect is needed
                    if response.status_code in [301, 302]:
                        redirect_url = response.headers.get('Location')
                        if redirect_url:
                            if debug:
                                log(f"Redirecting to: {redirect_url}")
                            # Retry the upload at the new endpoint
                            url = redirect_url
                            continue
                        else:
                            raise ValueError(
                                'Redirect location not provided in response')
                    else:
                        # If no redirect is needed or request is successful, break the loop
                        response.raise_for_status()
                        return response

                except requests.RequestException as e:
                    if debug:
                        log(f"Error during upload: {str(e)}")
                    if attempt < max_retries - 1:
                        if debug:
                            log("Retrying...")
                    else:
                        raise  # Re-raise the exception if max retries exceeded

        return upload_file_with_redirect_handling(
            presigned_post_data['url'], presigned_post_data['fields'], file_path)

    try:
        upload_response = run_step(
            f'Upload package to project:{project_name}', step2_upload_file, fail_mode=FailMode.THROW_ERROR)

        json = POST(platform_url, {
            'id': presigned_post_data['database_id'],
            'status': 'success'
        })
    except Exception as e:
        json = POST(platform_url, {
            'id': presigned_post_data['database_id'],
            'status': 'error',
            'error': serialize_exception(e)
        })


def push_project(project_name: str, path_to_code: str, debug: bool = False):
    """Push code to a project"""

    def step1_zip(log):
        if debug:
            log(f'path_to_code: {path_to_code}')
        path_to_zip = zip_directory_in_isolated_tempdir_v2(
            path_to_code)
        if debug:
            log(f'path_to_zip: {path_to_zip}')
        filename = os.path.basename(path_to_zip)
        return path_to_zip, filename

    path_to_zip, filename = run_step('Package code as .zip', step1_zip)

    __upload_file_via_presigned_url(
        path_to_zip, '/cli/push', project_name, debug)


def set_project_env(project_name: str, path_to_env: str, debug: bool = False):
    __upload_file_via_presigned_url(
        path_to_env, '/cli/set_env', project_name, debug)



import os
from typing import Optional
import yaml

from supadef.config import LOCAL_CREDS_PATH


def parse_credentials(path):
    text = open(path).read()
    creds = yaml.safe_load(text)
    return creds


def get_api_key(profile: Optional[str] = None):
    # 1. use the env var if it exists
    SUPADEF_API_KEY = os.getenv("SUPADEF_API_KEY", None)
    if SUPADEF_API_KEY:
        return SUPADEF_API_KEY
    # 2. otherwise, look for ~/.supadef/credentials.yml
    path = os.path.expanduser(LOCAL_CREDS_PATH)
    try:
        creds = parse_credentials(path)
    except FileNotFoundError as err:
        raise ValueError(
            f'No api key detected. Please set (1) SUPADEF_API_KEY or (2) {path}')

    if profile:
        profile_api_key = creds.get(profile)
        if profile_api_key:
            return profile_api_key
        raise ValueError(f"No profile named '{profile}'")

    api_key = creds.get('default')

    if not api_key:
        raise ValueError(
            'Please include your API Key by using the "default" attribute. Only the default key is supported, for now.')

    return api_key
